# User runs these tests with a `Upath` object of their subclass,
# in a path that is safe for testing.

import concurrent.futures
import os
import pathlib
import random
import time
from uuid import uuid4

import pytest

from upathlib import LocalUpath, LockAcquireError, Upath

IS_WIN = os.name != "posix"


def test_basic(p: Upath):
    pp = p / "/abc/def/"
    if isinstance(pp, LocalUpath):
        assert pp.path == pathlib.Path("/abc/def").absolute()
    else:
        assert pp.path == pathlib.PurePath("/abc/def")
    print(repr(pp))

    pp = pp / "x/y/z"
    if isinstance(pp, LocalUpath):
        assert pp.path == pathlib.Path("/abc/def/x/y/z").absolute()
    else:
        assert pp.path == pathlib.PurePath("/abc/def/x/y/z")

    print(repr(pp))

    pp /= "xy/z"
    if isinstance(pp, LocalUpath):
        assert str(pp.path) == str(pathlib.Path("/abc/def/x/y/z/xy/z").absolute())
    else:
        assert str(pp.path) == "/abc/def/x/y/z/xy/z"

    assert pp._path == str(pp.path)
    pp /= ".."
    if isinstance(pp, LocalUpath):
        assert pp._path == str(pathlib.Path("/abc/def/x/y/z/xy").absolute())
    else:
        assert pp._path == "/abc/def/x/y/z/xy"

    if isinstance(pp, LocalUpath):
        pp.joinpath("..")._path == str(pathlib.Path("/abc/def/x/y/z").absolute())
        pp.joinpath("..", "..", "..", "..", "..")._path == str(
            pathlib.Path("/").absolute()
        )
    else:
        pp.joinpath("..")._path == "/abc/def/x/y/z"
        pp.joinpath("..", "..", "..", "..", "..")._path == "/"


def test_joinpath(path: Upath):
    try:
        pp = path.joinpath("/abc/def/", "x/y") / "ab.txt"
        if isinstance(pp, LocalUpath):
            assert str(pp.path) == str(pathlib.Path("/abc/def/x/y/ab.txt").absolute())
        else:
            assert str(pp.path) == "/abc/def/x/y/ab.txt"

        pp = pp.joinpath("../a/b.txt")
        assert pp == path / "/abc/def" / "x/y/a/b.txt"
        assert pp.name == "b.txt"
        assert pp.suffix == ".txt"

        p = pp

        pp = pp / "../../../../"
        if isinstance(pp, LocalUpath):
            assert str(pp.path) == str(pathlib.Path("/abc/def").absolute())
        else:
            assert str(pp.path) == "/abc/def"

        pp = p.joinpath("a", ".", "b/c.data")
        if isinstance(pp, LocalUpath):
            assert str(pp.path) == str(
                pathlib.Path("/abc/def/x/y/a/b.txt/a/b/c.data").absolute()
            )
        else:
            assert str(pp.path) == "/abc/def/x/y/a/b.txt/a/b/c.data"

    except Exception:
        print("")
        print("repr:  ", repr(pp))
        print("str:   ", str(pp))
        print("path:  ", pp.path)
        print("_path: ", pp._path)
        raise


def test_compare(p: Upath):
    assert p.joinpath("abc/def") / "x/y/z" == p / "abc/def/x/y" / "z"
    assert p / "abc/def" < p.joinpath("abc/def", "x")
    assert p.joinpath("abc/def/x", "y/z") > p.joinpath("abc/def", "x/y")


def test_read_write_rm_navigate(p: Upath):
    init_path = p._path

    p.rmrf()

    p1 = p / "abc.txt"
    assert not p1.exists()
    p1.write_text("abc")

    assert p1.is_file()
    assert not p1.is_dir()

    assert p1.exists()
    assert p1.read_text() == "abc"

    with pytest.raises(FileExistsError):
        p1.write_text("abcd")

    p1.write_json({"data": "abcd"}, overwrite=True)  # type: ignore
    assert p1.read_json() == {"data": "abcd"}  # type: ignore

    p /= "a"
    if isinstance(p, LocalUpath):
        assert p._path == str(pathlib.Path(f"{init_path}/a").absolute())
    else:
        assert p._path == f"{init_path}/a"

    assert not p.is_file()
    assert not p.is_dir()
    assert not p.exists()

    time.sleep(0.3)
    # wait to ensure the next file as diff timestamp
    # from the first file.

    p2 = p.joinpath("x.data")
    p2.write_bytes(b"x")
    p /= ".."
    assert p._path == init_path
    assert p2 == p.joinpath("a", "x.data")
    assert p2.read_bytes() == b"x"

    p3 = p / "a"

    assert p.ls() == [p3, p1]

    assert sorted(p.riterdir()) == [p2, p1]
    assert p3.file_info() is None
    fi1 = p1.file_info()
    fi2 = p2.file_info()
    print("")
    print("p1:", fi1)
    print("p2:", fi2)
    print("")
    assert fi1.mtime < fi2.mtime  # type: ignore
    print("file 1 size:", fi1.size)  # type: ignore
    print("file 2 size:", fi2.size)  # type: ignore
    assert fi1.size > fi2.size  # type: ignore

    assert p3.is_dir()
    assert p3.remove_dir() == 1
    assert not p3.exists()
    assert not p2.exists()

    assert p3.remove_dir() == 0
    assert p.ls() == [p1]
    assert p1.remove_dir() == 0
    p1.remove_file()
    with pytest.raises(FileNotFoundError):
        p1.remove_file()
    assert p.rmrf() == 0


def test_copy(p: Upath):
    source = p
    source.rmrf()

    target = LocalUpath("/tmp/upath-test-target") / str(uuid4())
    target.rmrf()
    try:
        source_file = source / "testfile"
        source_file.write_text("abc", overwrite=True)

        source_file.copy_file(target)
        assert target.read_text() == "abc"

        with pytest.raises(FileNotFoundError if IS_WIN else NotADirectoryError):
            # cant' write to `target/'samplefile'`
            # because `target` is a file.
            source.copy_dir(target.joinpath("samplefile"))

        target.rmrf()
        p2 = target.joinpath("samplefile")
        source.copy_dir(p2)
        p3 = p2 / source_file.name
        assert target.ls() == [p2]
        assert p2.ls() == [p3]
        assert p3.read_text() == "abc"

        p1 = source / "a" / "b" / "c"
        assert p2.copy_dir(p1) == 1
        p4 = p1 / source_file.name
        assert p4.read_text() == "abc"

        assert p2.copy_dir(source / "a" / "b") == 1
        assert (source / "a" / "b" / source_file.name).read_text() == "abc"
    finally:
        target.rmrf()


def _access_in_mp(root: Upath, path: str, timeout):
    p = root / path
    t0 = time.perf_counter()
    try:
        with p.lock(timeout=timeout):
            return time.perf_counter() - t0
    except LockAcquireError:
        return t0 - time.perf_counter()


def test_lock1(p: Upath, timeout=None, wait=8):
    p.rmrf()
    pp = p / "testlock"
    with pp.lock(timeout=timeout):
        with concurrent.futures.ProcessPoolExecutor(1) as pool:
            t = pool.submit(_access_in_mp, p / "/", pp._path, wait)
            z = t.result()
            print("mp returned after", z, "seconds")
            # The work is not able to acquire the lock:
            assert z <= -(wait / 2)
    assert not pp.exists()


def _inc_in_mp(counter, idx):
    t0 = time.perf_counter()
    n = 0
    while time.perf_counter() - t0 < 10:
        with counter.with_suffix(".lock").lock():
            x = counter.read_text()
            print("x:", x, "worker", idx, flush=True)
            time.sleep(random.random() * 0.3)
            counter.write_text(str(int(x) + 1), overwrite=True)
            n += 1
            print("        worker", idx, n, flush=True)
        time.sleep(random.random() * 0.2)
    return idx, n


def test_lock2(p: Upath):
    p.rmrf()
    counter = p / "counter"
    counter.write_text("0")
    time.sleep(0.2)
    with concurrent.futures.ProcessPoolExecutor(30) as pool:
        tt = [pool.submit(_inc_in_mp, counter, i) for i in range(30)]
        results = [t.result() for t in tt]
        print("results:")
        for v in sorted(results):
            print(v)
        total1 = sum(v[1] for v in results)
        total2 = int(counter.read_text())
        print("")
        print(total1, total2)
        assert total1 == total2
    assert not counter.with_suffix(".lock").exists()


def test_lock(p: Upath):
    test_lock1(p)
    test_lock2(p)


def test_all(p: Upath):
    test_basic(p)
    test_joinpath(p)
    test_compare(p)

    test_read_write_rm_navigate(p)
    test_copy(p)

    # test_lock(p)



from __future__ import annotations

# Enable using `Upath` in type annotations in the code
# that defines this class.
# https://stackoverflow.com/a/49872353
# Will no longer be needed in Python 3.10.
import contextlib
import logging
import os
import time
from collections.abc import Iterator
from datetime import datetime, timezone
from io import BufferedReader, BytesIO, UnsupportedOperation
from typing import Optional

import google.auth
import requests
import requests.exceptions
from google import resumable_media
from google.api_core import exceptions
from google.api_core.retry import if_exception_type
from google.auth import exceptions as auth_exceptions
from google.cloud import storage
from google.cloud.storage.retry import DEFAULT_RETRY
from overrides import overrides
from typing_extensions import Self

from ._blob import BlobUpath, LocalPathType, _resolve_local_path
from ._upath import FileInfo, LockAcquireError, LockReleaseError, Upath

logger = logging.getLogger(__name__)


# 67108864 = 256 * 1024 * 256 = 64 MB
MEGABYTES32 = 33554432
MEGABYTES64 = 67108864
LARGE_FILE_SIZE = MEGABYTES64


RETRY_WRITE_ON_EXCEPTIONS = [
    exceptions.InternalServerError,
    exceptions.TooManyRequests,
    exceptions.ServiceUnavailable,
    exceptions.GatewayTimeout,
    requests.exceptions.ReadTimeout,
    requests.exceptions.ConnectionError,
    requests.exceptions.ChunkedEncodingError,
    auth_exceptions.TransportError,
]


def get_google_auth(
    project_id=None,
    credentials=None,
    *,
    scopes: list[str] = None,
    valid_for_seconds: int = 300,
):
    renewed = False
    if project_id is None or credentials is None:
        cred, pid = google.auth.default(
            scopes=scopes or ["https://www.googleapis.com/auth/cloud-platform"]
        )
        if credentials is None:
            credentials = cred
        if project_id is None:
            project_id = pid
        renewed = True

    if (
        not credentials.token
        or (credentials.expiry - datetime.utcnow()).total_seconds() < valid_for_seconds
    ):
        credentials.refresh(google.auth.transport.requests.Request())
        # One check shows that this token expires in one hour.
        renewed = True

    return project_id, credentials, renewed


class GcsBlobUpath(BlobUpath):
    """
    GcsBlobUpath implements the :class:`~upathlib.Upath` API for
    Google Cloud Storage.
    """

    _PROJECT_ID: str = None
    _CREDENTIALS: google.auth.credentials.Credentials = None
    _CLIENT: storage.Client = None
    # The `storage.Client` object is not pickle-able.
    # But if it is copied into another "forked" process, it will function properly.
    # Hence this is safe with multiprocessing, be it forked or spawned.
    # In a "spawned" process, this will start as None.

    _LOCK_EXPIRE_IN_SECONDS: int = 600
    # Things performed while holding a `lock` should finish within
    # this many seconds. If a worker tries but fails to acquire a lock on a file,
    # and finds the lock file has existed this long, it assumes the file
    # is "dead" because somehow the previous locking failed to delete the file properly,
    # and it will delete this lock file, and retry lock acquisition.
    #
    # Usually you don't need to customize this.

    @classmethod
    def _client(cls) -> storage.Client:
        """
        Return a client to the GCS service.

        If you have GCP account_info in a dict with these elements
        (not sure everything here is required)::

            'type': 'service_account',
            'project_id':
            'private_key_id':
            'private_key':
                '-----BEGIN PRIVATE KEY-----\\n'
                + private_key.encode('latin1').decode('unicode_escape')
                + '\\n-----END PRIVATE KEY-----\\n',
            'client_email':
            'client_id':
            'auth_uri': 'https://accounts.google.com/o/oauth2/auth',
            'token_uri': 'https://oauth2.googleapis.com/token',
            'auth_provider_x509_cert_url': 'https://www.googleapis.com/oauth2/v1/certs',
            'client_x509_cert_url': f"https://www.googleapis.com/robot/v1/metadata/x509/{client_email.replace('@', '%40')}"

        then ``credentials`` are obtained by

        ::

            google.oauth2.service_account.Credentials.from_service_account_info(
                account_info, scopes=['https://www.googleapis.com/auth/cloud-platform'])

        Code that runs on a GCP machine may be able to infer ``credentials`` and ``project_id``
        via `google.auth.default() <https://googleapis.dev/python/google-auth/latest/user-guide.html#application-default-credentials>`_.
        """
        cls._PROJECT_ID, cls._CREDENTIALS, renewed = get_google_auth(
            cls._PROJECT_ID, cls._CREDENTIALS
        )
        if cls._CLIENT is None or renewed:
            cls._CLIENT = storage.Client(
                project=cls._PROJECT_ID, credentials=cls._CREDENTIALS
            )
        return cls._CLIENT

    def __init__(
        self,
        *paths: str,
        bucket_name: str = None,
    ):
        """
        If ``bucket_name`` is ``None``, then ``*paths`` should be a single string
        starting with 'gs://<bucket-name>/'.

        If ``bucket_name`` is specified, then ``*paths`` specify path components
        under the root of the bucket.

        Examples
        --------
        These several calls are equivalent:

        >>> GcsBlobUpath('experiments', 'data', 'first.data', bucket_name='backup')
        GcsBlobUpath('gs://backup/experiments/data/first.data')
        >>> GcsBlobUpath('/experiments/data/first.data', bucket_name='backup')
        GcsBlobUpath('gs://backup/experiments/data/first.data')
        >>> GcsBlobUpath('gs://backup/experiments/data/first.data')
        GcsBlobUpath('gs://backup/experiments/data/first.data')
        >>> GcsBlobUpath('gs://backup', 'experiments', 'data/first.data')
        GcsBlobUpath('gs://backup/experiments/data/first.data')
        """
        if bucket_name is None:
            # The first arg must be like
            #   'gs://bucket-name'
            # or
            #   'gs://bucket-name/path...'

            p0 = paths[0]
            assert p0.startswith("gs://")
            p0 = p0[5:]
            k = p0.find("/")
            if k < 0:
                bucket_name = p0
                paths = paths[1:]
            else:
                bucket_name = p0[:k]
                p0 = p0[k:]
                paths = (p0, *paths[1:])

        super().__init__(*paths)
        assert bucket_name
        self.bucket_name = bucket_name
        self._bucket_ = None
        self._lock_count: int = 0
        self._generation = -1
        self._quiet_multidownload = True

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.as_uri()}')"

    def __str__(self) -> str:
        return self.as_uri()

    def __getstate__(self):
        # Customize pickle because `self._bucket_`
        # (when not None) can't be pickled.
        # the `service_account.Credentials` class object can be pickled.
        return (
            self.bucket_name,
            self._quiet_multidownload,
        ), super().__getstate__()

    def __setstate__(self, data):
        (self.bucket_name, self._quiet_multidownload), z1 = data
        self._bucket_ = None
        self._lock_count = 0
        self._generation = -1
        super().__setstate__(z1)

    def _bucket(self) -> storage.Bucket:
        """
        Return a Bucket object, via :meth:`client`.
        """
        if self._bucket_ is None:
            self._bucket_ = self._client().bucket(self.bucket_name)
        return self._bucket_

    def _blob(self) -> storage.Blob:
        """
        This constructs a Blob object irrespective of whether the blob
        exists in GCS.
        """
        return self._bucket().blob(self.blob_name)

    @overrides
    def as_uri(self) -> str:
        """
        Represent the path as a file URI, like 'gs://bucket-name/path/to/blob'.
        """
        return f"gs://{self.bucket_name}/{self._path.lstrip('/')}"

    @overrides
    def is_file(self) -> bool:
        """
        The result of this call is not cached, in case the object is modified anytime
        by other clients.
        """
        return self._blob().exists(self._client())

    @overrides
    def is_dir(self) -> bool:
        """
        If there is a dummy blob with name ``f"{self.name}/"``,
        this will return ``True``.
        This is the case after creating a "folder" on the GCP dashboard.
        In programatic use, it's recommended to avoid such situations so that
        ``is_dir()`` returns ``True`` if and only if there are blobs
        "under" the current path.
        """
        prefix = self.blob_name + "/"
        blobs = self._client().list_blobs(
            self._bucket(),
            prefix=prefix,
            max_results=1,
            page_size=1,
            fields="items(name),nextPageToken",
        )
        return len(list(blobs)) > 0

    @overrides
    def file_info(self) -> Optional[FileInfo]:
        """
        Return file info if the current path is a file;
        otherwise return ``None``.
        """
        b = self._blob()
        try:
            b.reload(client=self._client())
        except exceptions.NotFound:
            return None
        return FileInfo(
            ctime=b.time_created.timestamp(),
            mtime=b.updated.timestamp(),
            time_created=b.time_created,
            time_modified=b.updated,
            size=b.size,  # bytes
            details=b._properties,
        )
        # If an existing file is written to again using `write_...`,
        # then its `ctime` and `mtime` are both updated.
        # My experiments showed that `ctime` and `mtime` are equal.

    @property
    @overrides
    def root(self) -> GcsBlobUpath:
        """
        Return a new path representing the root of the same bucket.
        """
        obj = self.__class__(
            bucket_name=self.bucket_name,
        )
        obj._bucket_ = self._bucket()
        return obj

    def _write_from_buffer(
        self, file_obj, *, overwrite=False, content_type=None, size=None
    ):
        if self._path == "/":
            raise UnsupportedOperation("can not write to root as a blob", self)

        try:
            self._blob().upload_from_file(
                file_obj,
                content_type=content_type,
                size=size,
                client=self._client(),
                if_generation_match=None if overwrite else 0,
            )
            # TODO: set "create_time", 'update_time" to be the same
            # as the source local file?
            # Blob objects has methods `_set_properties`, `_patch_property`,
            # `patch`.
        except exceptions.PreconditionFailed:
            raise FileExistsError(self)

    def _write_bytes(self, data, **kwargs):
        b = BytesIO(data)
        b.seek(0)
        self._write_from_buffer(b, content_type="text/plain", size=len(data), **kwargs)

    @overrides
    def write_bytes(self, data: bytes | BufferedReader, *, overwrite=False):
        """
        Write bytes ``data`` to the current blob.

        In the usual case, ``data`` is bytes.
        The case where ``data`` is a ``BufferedReader`` object, such as an open file,
        is not well tested.
        """
        if isinstance(data, bytes):
            self._blob_rate_limit(self._write_bytes, data, overwrite=overwrite)
            return
        self._write_from_buffer(data, content_type="text/plain", overwrite=overwrite)

    def _read_into_buffer(self, file_obj):
        file_info = self.file_info()
        if not file_info:
            raise FileNotFoundError(self)
        file_size = file_info.size  # bytes
        if file_size <= LARGE_FILE_SIZE:
            try:
                self._blob().download_to_file(file_obj, client=self._client())
                return
            except exceptions.NotFound:
                raise FileNotFoundError(self)

        def _download(client, blob, start, end):
            buffer = BytesIO()
            try:
                blob.download_to_file(buffer, client=client, start=start, end=end)
            except exceptions.NotFound:
                raise FileNotFoundError(blob.name)
            # Both `start` and `end` are inclusive.
            # The very first `start` should be 0.
            buffer.seek(0)
            return buffer, end - start + 1

        def _do_download():
            client = self._client()
            blob = self._blob()
            k = 0
            p = 0
            while True:
                kk = min(k + LARGE_FILE_SIZE, file_size)
                p += 1
                yield (
                    _download,
                    (client, blob, k, kk - 1),
                    {},
                    f"part {p}",
                )
                k = kk
                if k >= file_size:
                    break

        for buf, k in self._run_in_executor(_do_download(), self._quiet_multidownload):
            n = file_obj.write(buf.getbuffer())
            if n != k:
                raise BufferError(
                    f"expecting to read {k} bytes; actually read {n} bytes"
                )
            buf.close()

    @overrides
    def read_bytes(self) -> bytes:
        """
        Return the content of the current blob as bytes.
        """
        buffer = BytesIO()
        self._read_into_buffer(buffer)
        return buffer.getvalue()

    def _blob_rate_limit(self, func, *args, **kwargs):
        # `func_name` is a create/update/delete function.
        # Google imposes rate limiting on such requests.
        # According to Google doc, https://cloud.google.com/storage/quotas,
        #   There is a write limit to the same object name. This limit is once per second.
        f = DEFAULT_RETRY.with_predicate(if_exception_type(*RETRY_WRITE_ON_EXCEPTIONS))(
            func
        )
        # Apply this inside the function so that user could add elements
        # to ``RETRY_WRITE_ON_EXCEPTIONS``.
        return f(*args, **kwargs)

    @overrides
    def _copy_file(self, target: Upath, *, overwrite=False) -> None:
        if isinstance(target, GcsBlobUpath):
            # https://cloud.google.com/storage/docs/copying-renaming-moving-objects
            try:
                self._bucket().copy_blob(
                    self._blob(),
                    target._bucket(),
                    target.blob_name,
                    client=self._client(),
                    if_generation_match=None if overwrite else 0,
                )
            except exceptions.NotFound:
                raise FileNotFoundError(self)
            except exceptions.PreconditionFailed:
                raise FileExistsError(target)
        else:
            super()._copy_file(target, overwrite=overwrite)

    @overrides
    def download_file(self, target: LocalPathType, *, overwrite=False) -> None:
        """
        Download the content of the current blob to ``target``.
        """
        target = _resolve_local_path(target)
        if target.is_file():
            if not overwrite:
                raise FileExistsError(target)
            target.remove_file()
        elif target.is_dir():
            raise IsADirectoryError(target)

        os.makedirs(str(target.parent), exist_ok=True)
        try:
            with open(target, "wb") as file_obj:
                # If `target` is an existing directory,
                # will raise `IsADirectoryError`.
                self._read_into_buffer(file_obj)
            updated = self._blob().updated
            if updated is not None:
                mtime = updated.timestamp()
                os.utime(target, (mtime, mtime))
        except resumable_media.DataCorruption:
            target.remove_file()
            raise

    @overrides
    def upload_file(self, source: LocalPathType, *, overwrite=False) -> None:
        """
        Upload the content of ``source`` to the current blob.
        """
        source = _resolve_local_path(source)
        filename = str(source)
        content_type = self._blob()._get_content_type(None, filename=filename)

        if self.is_file():
            if not overwrite:
                raise FileExistsError(self)
            self.remove_file()

        def _upload():
            with open(filename, "rb") as file_obj:
                total_bytes = os.fstat(file_obj.fileno()).st_size
                self._write_from_buffer(
                    file_obj,
                    size=total_bytes,
                    content_type=content_type,
                    overwrite=overwrite,
                )

        self._blob_rate_limit(_upload)

    @overrides
    def iterdir(self) -> Iterator[Self]:
        """
        Yield immediate children under the current dir.
        """
        # From Google doc:
        #
        # Lists all the blobs in the bucket that begin with the prefix.
        #
        # This can be used to list all blobs in a "folder", e.g. "public/".
        #
        # The delimiter argument can be used to restrict the results to only the
        # "files" in the given "folder". Without the delimiter, the entire tree under
        # the prefix is returned. For example, given these blobs:
        #
        #     a/1.txt
        #     a/b/2.txt
        #
        # If you specify prefix ='a/', without a delimiter, you'll get back:
        #
        #     a/1.txt
        #     a/b/2.txt
        #
        # However, if you specify prefix='a/' and delimiter='/', you'll get back
        # only the file directly under 'a/':
        #
        #     a/1.txt
        #
        # As part of the response, you'll also get back a blobs.prefixes entity
        # that lists the "subfolders" under `a/`:
        #
        #     a/b/
        #
        # Search "List the objects in a bucket using a prefix filter | Cloud Storage"
        #
        # You can "create folder" on the Google Cloud Storage dashboard. What it does
        # seems to create a dummy blob named with a '/' at the end and sized 0.
        # This case is handled in this function.

        prefix = self.blob_name + "/"
        k = len(prefix)
        blobs = self._client().list_blobs(self._bucket(), prefix=prefix, delimiter="/")
        for p in blobs:
            if p.name == prefix:
                # This happens if users has used the dashboard to "create a folder".
                # This seems to be a valid blob except its size is 0.
                # If user deliberately created a blob with this name and with content,
                # it's ignored. Do not use this name for a blob!
                continue
            obj = self / p.name[k:]  # files
            yield obj
        for p in blobs.prefixes:
            yield self / p[k:].rstrip("/")  # "subdirectories"
            # If this is an "empty subfolder", it is counted but it can be
            # misleading. User should avoid creating such empty folders.

    @overrides
    def remove_dir(self, **kwargs) -> int:
        """
        Remove the current dir and all the content under it recursively.
        Return the number of blobs removed.
        """
        z = super().remove_dir(**kwargs)
        prefix = self.blob_name + "/"
        for p in self._client().list_blobs(self._bucket(), prefix=prefix):
            assert p.name.endswith("/")
            p.delete()
        return z

    @overrides
    def remove_file(self) -> None:
        """
        Remove the current blob.
        """
        try:
            self._blob_rate_limit(self._blob().delete, client=self._client())
        except exceptions.NotFound:
            raise FileNotFoundError(self)

    @overrides
    def riterdir(self) -> Iterator[Self]:
        """
        Yield all blobs recursively under the current dir.
        """
        prefix = self.blob_name + "/"
        k = len(prefix)
        for p in self._client().list_blobs(self._bucket(), prefix=prefix):
            if p.name.endswith("/"):
                # This can be an "empty folder"---better not create them!
                # Worse, this is an actual blob name---do not do this!
                continue
            obj = self / p.name[k:]
            yield obj

    def _acquire_lease(self, *, timeout: int = None):
        # Note: `timeout = None` does not mean infinite wait.
        # It means a default wait time. If user wants longer wait,
        # just pass in a large number. Because user often associate
        # `timeout = None` with infinite wait, the default wait
        # is a long period.
        if self._path == "/":
            raise UnsupportedOperation("can not write to root as a blob", self)
        if timeout is None:
            timeout = 120  # seconds

        retry = DEFAULT_RETRY.with_timeout(timeout).with_predicate(
            if_exception_type(
                *RETRY_WRITE_ON_EXCEPTIONS,
                exceptions.PreconditionFailed,
                FileExistsError,
            )
        )

        @retry
        def _acquire_():
            self._write_bytes(b"0")
            self._generation = self._blob().generation

        t0 = time.perf_counter()
        try:
            _acquire_()
        except FileExistsError as e:
            finfo = self.file_info()
            now = datetime.utcnow().replace(tzinfo=timezone.utc)
            file_age = (now - finfo.time_created).total_seconds()
            if file_age - timeout > self._LOCK_EXPIRE_IN_SECONDS:
                # If the file is old,
                # assume it is a dead file, that is, the last lock operation
                # somehow failed and did not delete the file.
                logger.warning(
                    "the locker file '%s' was created %d seconds ago; assuming it is dead and deleting it",
                    self,
                    int(file_age),
                )
                self._blob_rate_limit(
                    self._blob().delete,
                    client=self._client(),
                )  # If this fails, the exception will propagate, which is not LockAcquireError.
                # After deleting the file, try it again:
                self._acquire_lease(timeout=timeout)
            else:
                raise LockAcquireError(
                    f"waited on '{self}' for {time.perf_counter() - t0:.2f} seconds"
                ) from e
        except Exception as e:
            raise LockAcquireError(
                f"waited on '{self}' for {time.perf_counter() - t0:.2f} seconds"
            ) from e

    @contextlib.contextmanager
    @overrides
    def lock(self, *, timeout: int | float = None):
        """
        This implementation does not prevent the file from being deleted
        by other workers that does not use the 'if-generation-match' condition.
        It relies on the assumption that this blob
        is used *cooperatively* solely in this locking logic.

        ``timeout`` is the wait time for acquiring the lease.
        If ``None``, the default value 100 seconds is used.
        If ``0``, exactly one attempt is made to acquire a lock.
        """
        # References:
        # https://www.joyfulbikeshedding.com/blog/2021-05-19-robust-distributed-locking-algorithm-based-on-google-cloud-storage.html
        # https://cloud.google.com/storage/docs/generations-preconditions
        # https://cloud.google.com/storage/docs/gsutil/addlhelp/ObjectVersioningandConcurrencyControl

        if self._lock_count == 0:
            self._acquire_lease(timeout=timeout)
        self._lock_count += 1
        try:
            yield
        finally:
            self._lock_count -= 1
            if self._lock_count == 0:
                try:
                    self._blob_rate_limit(
                        self._blob().delete,
                        client=self._client(),
                        if_generation_match=self._generation,  # TODO: should we remove this condition?
                    )
                except Exception as e:
                    raise LockReleaseError(f"failed to delete lock file {self}") from e

    def open(self, mode="r", **kwargs):
        """
        Use this on a blob (not a "directory") as a context manager.
        See Google documentation.
        """
        return self._blob().open(mode, **kwargs)



from __future__ import annotations

import pathlib
import sys
from collections.abc import Iterator

from overrides import EnforceOverrides, overrides
from typing_extensions import Self

from ._local import LocalPathType, LocalUpath
from ._upath import Upath


def _resolve_local_path(p: LocalPathType):
    if isinstance(p, str):
        p = pathlib.Path(p)
    if isinstance(p, pathlib.Path):
        p = LocalUpath(str(p.resolve().absolute()))
    else:
        assert isinstance(p, LocalUpath)
    return p


class BlobUpath(Upath, EnforceOverrides):
    """
    BlobUpath is a base class for paths in a *cloud* storage, aka "blob store".
    This is in contrast to a *local* disk storage, which is implemnted by :class:`LocalUpath`.
    """

    @property
    def blob_name(self) -> str:
        """
        Return the "name" of the blob. This is the "path" without a leading ``'/'``.
        In cloud blob stores, this is exactly the name of the blob. The name often
        contains ``'/'``, which has no special role in the name per se but is *interpreted*
        by users to be a directory separator.
        """
        return self._path.lstrip("/")

    @overrides
    def is_dir(self) -> bool:
        """In a typical blob store, there is no such concept as a
        "directory". Here we emulate the concept in a local file
        system. If there is a blob named like

        ::

            /ab/cd/ef/g.txt

        we say there exists directory "/ab/cd/ef".
        We should never have a trailing `/` in a blob's name, like

        ::

            /ab/cd/ef/

        (I don't know whether the blob stores allow
        such blob names.)

        Consequently, ``is_dir`` is equivalent
        to "having stuff in the dir". There is no such thing as
        an "empty directory" in blob stores.
        """
        try:
            next(self.iterdir())
            return True
        except StopIteration:
            return False

    @overrides
    def iterdir(self) -> Iterator[Self]:
        """
        Yield immediate children under the current dir.

        This is a naive, inefficient implementation.
        Expected to be refined by subclasses.
        """
        p0 = self._path  # this could be '/'.
        if not p0.endswith("/"):
            p0 += "/"
        np0 = len(p0)
        subdirs = set()
        for p in self.riterdir():
            tail = p._path[np0:]
            if tail.startswith("/"):
                raise Exception(f"malformed blob name '{p._path}'")
            if "/" in tail:
                tail = tail[: tail.find("/")]
            if tail not in subdirs:
                yield self / tail
                subdirs.add(tail)

    def download_dir(
        self, target: LocalPathType, *, overwrite: bool = False, quiet: bool = False
    ) -> int:
        """
        A specialization of :meth:`~upathlib.Upath.copy_dir` where the target location
        is on the local disk.
        """
        target = _resolve_local_path(target)

        if not quiet:
            print(f"Downloading from {self!r} into {target!r}", file=sys.stderr)
        return self._copy_dir(
            self, target, "download_file", overwrite=overwrite, quiet=quiet
        )

    def download_file(self, target: LocalPathType, *, overwrite=False) -> None:
        """
        A specialization of :meth:`~upathlib.Upath.copy_file` where the target location
        is on the local disk.

        Subclass is expected to override with a more efficient implementation.
        """
        self.copy_file(target, overwrite=overwrite)

    def upload_dir(
        self, source: LocalPathType, *, overwrite: bool = False, quiet: bool = False
    ) -> int:
        """
        A specialization of :meth:`~upathlib.Upath.copy_dir` for :class:`~upathlib.LocalUpath`
        where the target location is in a cloud blob store.
        """
        source = _resolve_local_path(source)
        if not quiet:
            print(f"Importing from {source!r} into {self!r}", file=sys.stderr)
        return self._copy_dir(
            source, self, "upload_file", overwrite=overwrite, quiet=quiet, reversed=True
        )

    def upload_file(self, source: LocalPathType, *, overwrite=False) -> None:
        """
        A specialization of :meth:`~upathlib.Upath.copy_file` for :class:`~upathlib.LocalUpath`
        where the target location is in a cloud blob store.

        Subclass is expected to override with a more efficient implementation.
        """
        source.copy_file(self, overwrite=overwrite)



"""
:class:`Upath` is an abstract base class that defines the APIs and some of the implementation.
Subclasses tailor to particular storage systems.
Currently there are two production-ready subclasses; they implement ``Upath``
for local file systems and Google Cloud Storage, respectively.

The APIs follow the style of the standard library
`pathlib <https://docs.python.org/3/library/pathlib.html>`_ where appropriate.
"""

from __future__ import annotations

# Enable using `Upath` in type annotations in the code
# that defines this class.
# https://stackoverflow.com/a/49872353
# Will no longer be needed in Python 3.10.
import abc
import contextlib
import datetime
import functools
import os
import os.path
import pathlib
import queue
import sys
import threading
from collections.abc import Iterable, Iterator
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from io import UnsupportedOperation
from typing import (
    Any,
    Callable,
    Optional,
)

from mpservice.util import MAX_THREADS, get_shared_thread_pool
from overrides import EnforceOverrides
from tqdm.auto import tqdm
from typing_extensions import Self

from .serializer import (
    JsonSerializer,
    OrjsonSerializer,
    PickleSerializer,
    ZOrjsonSerializer,
    ZPickleSerializer,
    ZstdOrjsonSerializer,
    ZstdPickleSerializer,
)

# End user may want to do this:
#  logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)
# to suppress the "urllib3 connection lost" warning.


class LockAcquireError(TimeoutError):
    pass


class LockReleaseError(RuntimeError):
    pass


@dataclass
class FileInfo:
    ctime: float  #: Creation time as a POSIX timetamp.
    mtime: float  #: Last modification time as a POSIX timestamp.
    time_created: datetime.datetime  #: Creation time as an ``datetime`` object.
    time_modified: datetime.datetime  #: Last modification time as an ``datetime`` object.
    size: int  #: In bytes.
    details: Any  #: Platform-dependent.


@functools.total_ordering
class Upath(abc.ABC, EnforceOverrides):
    def __init__(
        self,
        *pathsegments: str,
    ):
        """
        Create a ``Upath`` instance. Because ``Upath`` is an abstract class,
        this is always called on a subclass to instantiate a path on the specific
        storage system.

        Subclasses for cloud blob stores may need to add additional parameters
        representing, e.g., container/bucket name, etc.

        Parameters
        ----------
        *pathsegments
            Analogous to the input to `pathlib.Path <https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_.
            The first segment may or may not start with ``'/'``.
            The path constructed with ``*pathsegments`` is always "absolute" under a known "root".

            For a local POSIX file system, the root is the usual ``'/'``.

            For a local Windows file system, the root is resolved to a particular "drive".

            For Azure blob store, the root is that in a "container".

            For AWS and GCP blob stores, the root is that in a "bucket".

            If missing, the path constructed is the "root". However,
            the subclass :class:`LocalUpath` plugs in the current working directory
            for a missing ``*pathsegments``.

            .. note:: If one segment starts with ``'/'``, it will reset to the "root"
                and discard all the segments that have come before it. For example,
                ``Upath('work', 'projects', '/', 'projects')``
                is the same as ``Upath('/', 'projects)``.

            .. note:: The first element of ``*pathsegments`` may start with some platform-specific
                strings. For example, ``'/'`` on Linux, ``'c://'`` on Windows, ``'gs://'`` on
                Google Cloud Storage. Please see subclasses for specifics.
        """

        self._path = os.path.normpath(
            os.path.join("/", *pathsegments)
        )  # pylint: disable=no-value-for-parameter
        # For LocalUpath on Windows, this is like 'C:\\Users\\username\\path'.
        # For LocalUpath on Linux, and BlobUpath, this is always absolute starting with '/'.
        # It does not have a trailing `/` unless the path is just `/` itself.
        self._thread_pools = {}

    def __getstate__(self):
        # Do not pickle `self._thread_pools`.
        return (self._path,)

    def __setstate__(self, data):
        self._path = data[0]
        self._thread_pools = {}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self._path}')"
        # Subclass may want to customize this method to add more info,
        # e.g. "bucket" name.

    def __str__(self) -> str:
        return self._path

    def __eq__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self.as_uri() == other.as_uri()

    def __lt__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self.as_uri() < other.as_uri()

    def __hash__(self) -> int:
        return hash(self.as_uri())

    def __truediv__(self, key: str) -> Self:
        """
        This method is invoked by ``self / key``.
        This calls the method :meth:`joinpath`.
        """
        return self.joinpath(key)

    def _get_thread_pool(self, name: str) -> ThreadPoolExecutor:
        # Currently there can be two "layers" of threads running during `download_dir`.
        # In `download_dir`, the download of each file runs in the threads provided
        # by `UpathExecutor0`. If a file is large, `GcsBlobUpath` will split the work into chunks
        # and download each chunk in a thread provided by `UpathExecutor1`.
        # We dedicate the two executors mainly so that the second layer of chunk downloads
        # do not starve for threads.
        if name == "first":
            return get_shared_thread_pool("upathlib-first", MAX_THREADS)
        elif name == "second":
            return get_shared_thread_pool("upathlib-second", MAX_THREADS - 2)
        else:
            raise ValueError(name)

    def _run_in_executor(
        self,
        tasks: Iterable[tuple[Callable, tuple, dict, str]],
        quiet: bool,
    ):
        """
        This method is used to run multiple I/O jobs concurrently, e.g.
        uploading/downloading all files in a folder recursively.

        Parameters
        ----------
        tasks
            Each element is a tuple of (func, args, kwargs, description).
        """
        if not isinstance(tasks, list):
            tasks = list(tasks)
        n_tasks = len(tasks)
        if not n_tasks:
            return

        pbar = None
        if threading.current_thread().name.startswith("UpathExecutor0"):
            # This is the case when GCP downloads a large file by multiple parts.
            # This is working on one large file, and it is trying to start threads
            # to tackle parts of it.
            executor = self._get_thread_pool("second")
            # No progress printout.
        else:
            executor = self._get_thread_pool("first")
            if not quiet:
                pbar = tqdm(
                    total=n_tasks,
                    bar_format="{percentage:5.1f}%, {n:.0f}/{total_fmt}, {elapsed} | {desc}",
                )

        def enqueue(tasks, executor, q, to_stop):
            for func, args, kwargs, desc in tasks:
                t = executor.submit(func, *args, **kwargs)
                # This has limited capacity, to control the speed of
                # task submission to the executor.
                q.put((t, desc))
                if to_stop.is_set():
                    break
            q.put(None)

        try:
            q = queue.Queue(executor._max_workers)
            to_stop = threading.Event()
            task = executor.submit(enqueue, tasks, executor, q, to_stop)

            try:
                while True:
                    z = q.get()
                    if z is None:
                        break
                    t, desc = z
                    if pbar:
                        pbar.set_description_str(desc)
                        pbar.update(0.5)
                    try:
                        yield t.result()
                    except Exception:
                        to_stop.set()
                        while True:
                            z = q.get()
                            if z is None:
                                break
                            t, desc = z
                            t.cancel()
                            # This may not succeed, but there isn't a good way to
                            # guarantee cancellation here.
                        raise
                    if pbar:
                        pbar.update(0.5)
            finally:
                _ = task.result()
        finally:
            if pbar:
                pbar.close()

    @property
    def path(self) -> pathlib.PurePath:
        """
        The `pathlib.PurePath <https://docs.python.org/3/library/pathlib.html#pathlib.PurePath>`_
        version of the internal path string.

        In the subclass :class:`LocalUpath`, this property is overriden to return a
        `pathlib.Path <https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_,
        which is a subclass of ``pathlib.PurePath``.

        In subclasses for cloud blob stores, this implementation stays in effect.
        """
        return pathlib.PurePath(self._path)

    @abc.abstractmethod
    def as_uri(self) -> str:
        """
        Represent the path as a file URI.
        See subclasses for platform-dependent specifics.
        """
        raise NotImplementedError

    @property
    def name(self) -> str:
        """
        A string representing the final path component, excluding the drive and root, if any.

        This is the ``name`` component of ``self.path``.
        If ``self.path`` is ``'/'`` (the root), then an empty string is returned.
        (The name of the root path is empty.)

        Examples
        --------
        >>> from upathlib import LocalUpath
        >>> p = LocalUpath('/tmp/test/upathlib/data/sales.txt.gz')
        >>> p.path
        PosixPath('/tmp/test/upathlib/data/sales.txt.gz')
        >>> p.name
        'sales.txt.gz'
        >>> p.parent.parent.parent.parent
        LocalUpath('/tmp')
        >>> p.parent.parent.parent.parent.name
        'tmp'
        >>> p.parent.parent.parent.parent.parent
        LocalUpath('/')
        >>> p.parent.parent.parent.parent.parent.name
        ''
        >>> # the parent of root is still root:
        >>> p.parent.parent.parent.parent.parent.parent
        LocalUpath('/')
        """
        return self.path.name

    @property
    def stem(self) -> str:
        """
        The final path component, without its suffix.

        This is the "stem" part of ``self.name``.

        Examples
        --------
        >>> from upathlib import LocalUpath
        >>> p = LocalUpath('/tmp/test/upathlib/data/sales.txt')
        >>> p
        LocalUpath('/tmp/test/upathlib/data/sales.txt')
        >>> p.path
        PosixPath('/tmp/test/upathlib/data/sales.txt')
        >>> p.name
        'sales.txt'
        >>> p.stem
        'sales'
        >>> p = LocalUpath('/tmp/test/upathlib/data/sales.txt.gz')
        >>> p.stem
        'sales.txt'
        """
        return self.path.stem

    @property
    def suffix(self) -> str:
        """
        The file extension of the final component, if any
        """
        return self.path.suffix

    @property
    def suffixes(self) -> list[str]:
        """
        A list of the path's file extensions.

        Examples
        --------
        >>> p = LocalUpath('/tmp/test/upathlib/data/sales.txt')
        >>> p.suffix
        '.txt'
        >>> p.suffixes
        ['.txt']
        >>> p = LocalUpath('/tmp/test/upathlib/data/sales.txt.gz')
        >>> p.suffix
        '.gz'
        >>> p.suffixes
        ['.txt', '.gz']
        """
        return self.path.suffixes

    def exists(self) -> bool:
        """Return ``True`` if the path is an existing file or dir;
        ``False`` otherwise.

        Examples
        --------
        In a blobstore with blobs

        ::

            /a/b/cd
            /a/b/cd/e.txt

        ``'/a/b/cd'`` exists, and is both a file and a dir;
        ``'/a/b/cd/e.txt'`` exists, and is a file;
        ``'/a/b'`` exists, and is a dir;
        ``'/a/b/c'`` does not exist.
        """
        return self.is_file() or self.is_dir()

    @abc.abstractmethod
    def is_dir(self) -> bool:
        """Return ``True`` if the path is an existing directory; ``False`` otherwise.

        If there exists a file named like

        ::

            /a/b/c/d.txt

        we say ``'/a'``, ``'/a/b'``, ``'/a/b/c'`` are existing directories.

        In a cloud blob store, there's no such thing as an
        "empty directory", because there is no concept of "directory".
        A blob store just consists of files (aka blobs) with names,
        which could contain the letter '/', with no special meaning
        attached to it.
        We interpret the name ``'/a/b'`` as a directory
        to emulate the familiar concept in a local file system when
        there exist files named ``'/a/b/*'``.

        In a local file system, there can be empty directories.
        However, it is recommended to not have empty directories.

        There is no method for "creating an tempty dir" (like ``mkdir``).
        Simply create a file under the dir, and the dir will come into being.
        This is analogous to we create files all the time---we don't "create" an empty file
        in advance; we simply write to the would-be path of the file to be created.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_file(self) -> bool:
        """Return ``True`` if the path is an existing file; ``False`` otherwise.

        In a cloud blob store, a path can be both a file and a dir. For
        example, if these blobs exist::

            /a/b/c/d.txt
            /a/b/c

        we say ``/a/b/c`` is a "file", and also a "dir".
        User is recommended to avoid such namings.

        This situation does not happen in a local file system.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def file_info(self) -> Optional[FileInfo]:
        """
        If :meth:`is_file` is ``False``, return ``None``; otherwise, return file info.
        """
        raise NotImplementedError

    @property
    def parent(self) -> Self:
        """
        The parent of the path.

        If the path is the root, then the parent is still the root.
        """
        return self._with_path(str(self.path.parent))

    @property
    @abc.abstractmethod
    def root(self) -> Self:
        """
        Return a new path representing the root.
        """
        raise NotImplementedError

    def _with_path(self, *paths) -> Self:
        """
        Return a new object of the same class at the specified ``*paths``.
        The new path is unrelated to the current path; in other words,
        the new path is not "relative" to the current path.

        The main use case is with a cloud blob store.
        For example, return a new path in the same store with the same
        account and bucket info.
        """
        # TODO: the implementation is a little hacky.
        r = self.root
        r._path = os.path.normpath(os.path.join("/", *paths))
        return r

    def joinpath(self, *other: str) -> Self:
        """Join this path with more segments, return the new path object.

        Calling this method is equivalent to combining the path with each
        of the ``other`` arguments in turn.

        If ``self`` was created by ``Upath(*segs)``, then this method essentially
        returns ``Upath(*segs, *other)``.

        If ``*other`` is a single string, there is a shortcut by the operator
        ``/``, implemented by :meth:`__truediv__`.
        """
        return self._with_path(os.path.join(self._path, *other))

    def with_name(self, name: str) -> Self:
        """
        Return a new path the the "name" part substituted by the new value.
        If the original path doesn't have a name (i.e. the original path is the root),
        ``ValueError`` is raised.

        Examples
        --------
        >>> p = LocalUpath('/tmp/test/upathlib/data/sales.txt.gz')
        >>> p.with_name('sales.data')
        LocalUpath('/tmp/test/upathlib/data/sales.data')
        """
        return self._with_path(str(self.path.with_name(name)))

    # def with_stem(self, stem: str) -> Self:
    #     # Available in Python 3.9+.
    #     return self._with_path(str(self.path.with_stem(stem)))

    def with_suffix(self, suffix: str) -> Self:
        """
        Return a new path with the suffix replaced by the specified value.
        If the original path doesn't have a suffix, the new suffix is appended instead.
        If ``suffix`` is an empty string, the original suffix is removed.

        ``suffix`` should include a dot, like ``'.txt'``.

        Examples
        --------
        >>> p = LocalUpath('/tmp/test/upathlib/data/sales.txt.gz')
        >>>
        >>> # replace the last suffix:
        >>> p.with_suffix('.data')
        LocalUpath('/tmp/test/upathlib/data/sales.txt.data')
        >>>
        >>> # remove the last suffix:
        >>> p.with_suffix('')
        LocalUpath('/tmp/test/upathlib/data/sales.txt')
        >>>
        >>> p.with_suffix('').with_suffix('.bin')
        LocalUpath('/tmp/test/upathlib/data/sales.bin')
        >>>
        >>> pp = p.with_suffix('').with_suffix('')
        >>> pp
        LocalUpath('/tmp/test/upathlib/data/sales')
        >>>
        >>> # no suffix to remove:
        >>> pp.with_suffix('')
        LocalUpath('/tmp/test/upathlib/data/sales')
        >>>
        >>> # add a suffix:
        >>> pp.with_suffix('.pickle')
        LocalUpath('/tmp/test/upathlib/data/sales.pickle')
        """
        return self._with_path(str(self.path.with_suffix(suffix)))

    @abc.abstractmethod
    def write_bytes(self, data: bytes, *, overwrite: bool = False) -> None:
        """Write bytes ``data`` to the current file.

        Parent "directories" are created as needed, if applicable.

        If ``overwrite`` is ``False`` and the current file exists, ``FileExistsError`` is raised.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def read_bytes(self) -> bytes:
        """
        Return the binary contents of the pointed-to file as a bytes object.

        If ``self`` is not a file or does not exist,
        ``FileNotFoundError`` is raised.
        """
        raise NotImplementedError

    def write_text(
        self,
        data: str,
        *,
        overwrite: bool = False,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
    ) -> None:
        """Write text ``data`` to the current file.

        Parent "directories" are created as needed, if applicable.

        If ``overwrite`` is ``False`` and the current file exists, ``FileExistsError`` is raised.

        ``encoding`` and ``errors`` are passed to `encode() <https://docs.python.org/3/library/stdtypes.html#str.encode>`_.
        Usually you should leave them at the default values.
        """
        z = data.encode(encoding=encoding or "utf-8", errors=errors or "strict")
        self.write_bytes(z, overwrite=overwrite)

    def read_text(
        self, *, encoding: Optional[str] = None, errors: Optional[str] = None
    ) -> str:
        """
        Return the decoded contents of the pointed-to file as a string.

        If ``self`` is not a file or does not exist,
        ``FileNotFoundError`` is raised.

        ``encoding`` and ``errors`` are passed to `decode() <https://docs.python.org/3/library/stdtypes.html#bytes.decode>`_.
        Usually you should leave them at the default values.
        """
        # Refer to https://docs.python.org/3/library/functions.html#open
        # and https://docs.python.org/3/library/codecs.html#module-codecs
        return self.read_bytes().decode(
            encoding=encoding or "utf-8", errors=errors or "strict"
        )

    def write_json(
        self, data: Any, *, encoding=None, errors=None, overwrite=False, **kwargs
    ) -> None:
        """
        ``encoding`` and ``errors`` are passed to :meth:`write_text`.

        ``overwrite`` is passed to :meth:`write_text`.

        ``**kwargs`` are passed to :meth:`serializer.JsonSerializer.serialize`.
        """
        self.write_text(
            JsonSerializer.serialize(data, **kwargs),
            overwrite=overwrite,
            encoding=encoding,
            errors=errors,
        )

    def read_json(self, *, encoding=None, errors=None, **kwargs) -> Any:
        """
        ``encoding`` and ``errors`` are passed to :meth:`read_text`.

        ``**kwargs`` are passed to :meth:`serializer.JsonSerializer.deserialize`.
        """
        return JsonSerializer.deserialize(
            self.read_text(encoding=encoding, errors=errors), **kwargs
        )

    def write_pickle(self, data: Any, *, overwrite=False, **kwargs) -> None:
        """
        ``overwrite`` is passed to :meth:`write_bytes`.

        ``**kwargs`` are passed to :meth:`serializer.PickleSerializer.serialize`.
        """
        self.write_bytes(
            PickleSerializer.serialize(data, **kwargs), overwrite=overwrite
        )

    def read_pickle(self, **kwargs) -> Any:
        """
        ``**kwargs`` are passed to :meth:`serializer.PickleSerializer.deserialize`.
        """
        return PickleSerializer.deserialize(self.read_bytes(), **kwargs)

    def write_pickle_z(self, data: Any, *, overwrite=False, **kwargs) -> None:
        """
        ``overwrite`` is passed to :meth:`write_bytes`.

        ``**kwargs`` are passed to :meth:`serializer.ZPickleSerializer.serialize`.
        """
        self.write_bytes(
            ZPickleSerializer.serialize(data, **kwargs), overwrite=overwrite
        )

    def read_pickle_z(self, **kwargs) -> Any:
        """
        ``**kwargs`` are passed to :meth:`serializer.ZPickleSerializer.deserialize`.
        """
        return ZPickleSerializer.deserialize(self.read_bytes(), **kwargs)

    def write_pickle_zstd(self, data: Any, *, overwrite=False, **kwargs) -> None:
        """
        ``overwrite`` is passed to :meth:`write_bytes`.

        ``**kwargs`` are passed to :meth:`serializer.ZstdPickleSerializer.serialize`.
        """
        self.write_bytes(
            ZstdPickleSerializer.serialize(data, **kwargs), overwrite=overwrite
        )

    def read_pickle_zstd(self, **kwargs) -> Any:
        """
        ``**kwargs`` are passed to :meth:`serializer.ZstdPickleSerializer.deserialize`.
        """
        return ZstdPickleSerializer.deserialize(self.read_bytes(), **kwargs)

    def write_orjson(self, data: Any, *, overwrite=False, **kwargs) -> None:
        """
        ``overwrite`` is passed to :meth:`write_bytes`.

        ``**kwargs`` are passed to :meth:`serializer.OrjsonSerializer.serialize`.
        """
        self.write_bytes(
            OrjsonSerializer.serialize(data, **kwargs), overwrite=overwrite
        )

    def read_orjson(self, **kwargs) -> Any:
        """
        ``**kwargs`` are passed to :meth:`serializer.OrjsonSerializer.deserialize`.
        """
        return OrjsonSerializer.deserialize(self.read_bytes(), **kwargs)

    def write_orjson_z(self, data: Any, *, overwrite=False, **kwargs) -> None:
        """
        ``overwrite`` is passed to :meth:`write_bytes`.

        ``**kwargs`` are passed to :meth:`serializer.ZOrjsonSerializer.serialize`.
        """
        self.write_bytes(
            ZOrjsonSerializer.serialize(data, **kwargs), overwrite=overwrite
        )

    def read_orjson_z(self, **kwargs) -> Any:
        """
        ``**kwargs`` are passed to :meth:`serializer.ZOrjsonSerializer.deserialize`.
        """
        return ZOrjsonSerializer.deserialize(self.read_bytes(), **kwargs)

    def write_orjson_zstd(self, data: Any, *, overwrite=False, **kwargs) -> None:
        """
        ``overwrite`` is passed to :meth:`write_bytes`.

        ``**kwargs`` are passed to :meth:`serializer.ZstdOrjsonSerializer.serialize`.
        """
        self.write_bytes(
            ZstdOrjsonSerializer.serialize(data, **kwargs), overwrite=overwrite
        )

    def read_orjson_zstd(self, **kwargs) -> Any:
        """
        ``**kwargs`` are passed to :meth:`serializer.ZstdOrjsonSerializer.deserialize`.
        """
        return ZstdOrjsonSerializer.deserialize(self.read_bytes(), **kwargs)

    def _copy_dir(
        self, source, dest, method: str, *, overwrite: bool, quiet: bool, reversed=False
    ):
        def foo():
            source_path = source.path
            ovwt = overwrite
            for p in source.riterdir():
                extra = str(p.path.relative_to(source_path))
                if reversed:
                    yield (
                        getattr(dest / extra, method),
                        (p,),
                        {"overwrite": ovwt},
                        extra,
                    )
                else:
                    yield (
                        getattr(p, method),
                        (dest / extra,),
                        {"overwrite": ovwt},
                        extra,
                    )

        n = 0
        for _ in self._run_in_executor(foo(), quiet):
            n += 1
        return n

    def copy_dir(
        self,
        target: str | Upath,
        *,
        overwrite: bool = False,
        quiet: bool = False,
    ) -> int:
        """Copy the content of the current directory (i.e. ``self``) recursively to ``target``.

        If ``target`` is an string, then it is in the same store the the current path,
        and it is either absolute, or relative to ``self.parent``.
        In this case,
        the directory created by this operation will be the path ``self.parent / target``.

        If ``target`` is not a string, then it must be an instance of a :meth:`Upath` subclass,
        and it may be in any store system.

        Immediate children of ``self`` will be copied as immediate children of the target path.

        There is no such error as "target directory exists" as the copy-operation
        only concerns invidivual files.
        If the target "directory" contains files that do not have counterparts
        in the source directory, they will stay untouched.

        ``overwrite`` is file-wise. If ``False``, any existing target file will raise ``FileExistsError`` and
        halt the operation. If ``True``, any existing target file will be overwritten by the source file.

        ``quiet`` controls whether to print out progress info.

        Returns
        -------
        int
            The number of files copied.
        """
        if isinstance(target, str):
            target_ = self.parent / target
            if target_ == self:
                return 0
        else:
            target_ = target

        if not quiet:
            print(f"Copying from {self!r} into {target_!r}", file=sys.stderr)
        return self._copy_dir(
            self, target_, "copy_file", overwrite=overwrite, quiet=quiet
        )

    def _copy_file(self, target: Upath, *, overwrite: bool = False) -> None:
        target.write_bytes(self.read_bytes(), overwrite=overwrite)

    def copy_file(self, target: str | Upath, *, overwrite: bool = False) -> None:
        """Copy the current file (i.e. ``self``) to ``target``.

        If ``target`` is str, then it is in the same store as the current path,
        and it is either absolute, or relative to ``self.parent``.
        In this case, the file created by this operation will the path ``self.parent / target``.
        For example, if ``self`` is ``'/a/b/c/d.txt'``, then
        ``target='e.txt'`` means ``'/a/b/c/e.txt'``.

        If ``target`` is not a string, then it must be an instance of a :meth:`Upath` subclass,
        and it may be in any storage system.

        ``target`` is the target file, *not* a target directory to "copy into".

        If ``self`` is not an existing file, ``FileNotFoundError`` is raised.

        If ``target`` is an existing file and ``overwrite`` is ``False``,
        ``FileExistsError`` is raised. If ``overwrite`` is ``True``,
        then the file will be overwritten.

        If ``type(self)`` is ``LocalUpath`` and ``target`` is an existing directory,
        then ``IsADirectoryError`` is raised. In a cloud blob store, there is no concrete "directory".
        For example, suppose ``self`` is the path 'gs://mybucket/experiment/data' on
        Google Cloud Storage, and ``target`` is '/backup/record', then
        the target path is 'gs://mybucket/backup/record'.
        If there exists blob 'gs://mybucket/backup/record/y', then we say 'gs://mybucket/backup/record'
        is a "directory". However, this is merely a "virtual" concept, or an emulation
        of the "directory" concept on local disk. As long as this path is not an
        existing blob, the copy will proceed with no problem.
        Nevertheless, such naming is confusing and better avoided.
        """
        if isinstance(target, str):
            target_ = self.parent / target
            if target_ == self:
                return
        else:
            target_ = target

        self._copy_file(target_, overwrite=overwrite)

    def remove_dir(self, *, quiet: bool = True) -> int:
        """Remove the current directory (i.e. ``self``) and all its contents recursively.

        Essentially, this removes each file that is yielded by :meth:`riterdir`.
        Subclasses should take care to remove "empty directories", if applicable,
        that are left behind.

        ``quiet`` controls whether to print progress info.

        Returns
        -------
        int
            The number of files removed.
        """

        def foo():
            for p in self.riterdir():
                yield p.remove_file, [], {}, str(p.path.relative_to(self.path))

        n = 0
        for _ in self._run_in_executor(foo(), quiet):
            n += 1
        return n

    @abc.abstractmethod
    def remove_file(self) -> None:
        """Remove the current file (i.e. ``self``).

        If ``self`` is not an existing file, ``FileNotFoundError`` is raised.
        If the file exists but can't be removed, the platform-dependent
        exception is propagated.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def iterdir(self) -> Iterator[Self]:
        """Yield the immediate (i.e. non-recursive) children
        of the current dir (i.e. ``self``).

        The yieled elements are instances of the same class.
        Each yielded element is either a file or a dir.
        There is no guarantee on the order of the returned elements.

        If ``self`` is not a dir (e.g. maybe it's a file),
        or does not exist at all, nothing is yielded (resulting in an
        empty iterable); no exception is raised.

        .. seealso:: :meth:`riterdir`.
        """
        raise NotImplementedError

    def ls(self) -> list[Self]:
        """Return the elements yielded by :meth:`iterdir` in a sorted list.

        Sorting is by a full path string maintained internally.

        The returned list may be empty.
        """
        return sorted(self.iterdir())

    @abc.abstractmethod
    def riterdir(self) -> Iterator[Self]:
        """Yield files under the current dir (i.e. ``self``) *recursively*.
        The method name means "recursive iterdir".

        The yieled elements are instances of the same class.
        They represent existing files.

        Compared to :meth:`iterdir`, this is recursive, and yields
        *files* only. Empty subdirectories will have no representation
        in the return.

        Similar to :meth:`iterdir`, if ``self`` is not a dir or does not exist,
        then nothing is yielded, and no exception is raised either.

        There is no guarantee on the order of the returned elements.
        """
        raise NotImplementedError

    def rmrf(self, *, quiet: bool = True) -> int:
        """Remove the current file or dir (i.e. ``self``) recursively.

        Analogous to the Linux command ``rm -rf``, hence the name of this method.

        Return the number of files removed.

        For example, if these blobs are present::

            /a/b/c/d/e.txt
            /a/b/c/kk.data
            /a/b/c

        then ``Upath('/a/b/c').rmrf()`` would remove all of them.
        """
        if self._path == "/":
            raise UnsupportedOperation("`rmrf` not allowed on root directory")
        try:
            self.remove_file()
        except (FileNotFoundError, IsADirectoryError):
            n = 0
        else:
            n = 1
        try:
            m = self.remove_dir(quiet=quiet)
        except FileNotFoundError:
            m = 0
        return n + m

    @contextlib.contextmanager
    @abc.abstractmethod
    def lock(self, *, timeout: int = None):
        """Lock the current file (i.e. ``self``), in order to have exclusive access to the code block
        that has possesion of the lock.

        ``timeout``: if the lock can't be acquired within ``timeout`` seconds,
        ``LockAcquireError`` is raised. If ``None``, wait for a default
        reasonably long time. To wait "forever", just pass in a large number.

        ``timeout=0`` is a valid input, meaning making exactly one attempt to acquire a lock.

        Once a lock is acquired, it will not expire until this contexmanager exits.
        In other words, this is timeout for the "lock acquisition", not for the
        lock itself. Actual waiting time could be slightly longer or shorter.

        This is a "mandatory lock", as opposed to an "advisory lock".
        The intended use case is for this lock to be used
        for implementing a (cooperative) "code lock".

        As this abstract method is to be used as a context manager,
        a subclass should use ``yield`` in its implementation.
        The ``yield`` statement is not required to yield anything,
        that is, it may be simply

        ::

            yield

        rather than, say,

        ::

            yield self

        One way to achive cooperative locking on a file via this
        lock is like this::

            f = Upath('abc.txt')
            with f.with_suffix('.txt.lock').lock():
                ...
                # Now write to `f` with exclusive access,
                # because any other (cooperative) code block
                # will not be able to get hold of `abc.txt.lock`
                # in order to write to `f` using this same code block
                # (or in another block that deliberately uses the same file lock).
                # Reading can be made exclusive by this same lock mechanism.

        .. note:: This method should always be called on an "auxiliary file" (like in the example above),
          whose sole purpose is to implement the lock; the file being locked should not be used for any other purpose.

        .. note:: Just before exiting the context manager, the implementation should
          delete the lock file.

        Some storage engines may not provide the capability to implement
        this lock.
        """
        raise NotImplementedError



from __future__ import annotations

# Enable using `Upath` in type annotations in the code
# that defines this class.
# https://stackoverflow.com/a/49872353
# Will no longer be needed in Python 3.10.
import os
import random
import time
from collections.abc import Iterator
from contextlib import contextmanager
from datetime import datetime
from io import UnsupportedOperation

# from azure.storage.blob.aio import (
# ContainerClient as aContainerClient,
# BlobClient as aBlobClient,
# BlobLeaseClient as aBlobLeaseClient,
# )
from azure.core.exceptions import (
    HttpResponseError,
    ResourceExistsError,
    ResourceNotFoundError,
)
from azure.storage.blob import BlobClient, BlobLeaseClient, ContainerClient
from overrides import overrides
from typing_extensions import Self

from ._blob import BlobUpath, LocalPathType, _resolve_local_path
from ._upath import FileInfo, LockAcquireError, LockReleaseError, Upath

# End user may want to do this:
# logging.getLogger("azure.storage").setLevel(logging.WARNING)
# logging.getLogger("azure.core.pipeline.policies").setLevel(logging.WARNING)


class AzureBlobUpath(BlobUpath):
    _ACCOUNT_NAME = None
    _ACCOUNT_KEY = None
    _SAS_TOKEN = None

    @classmethod
    def get_account_info(cls):
        # Subclass needs to customize this method or
        # hard-code relevant class attributes directly.
        # TODO: does Azure have a way to infer this info if the code
        # is running on an Azure machine?
        return {
            "account_url": f"https://{cls._ACCOUNT_NAME}.blob.core.windows.net",
            "credential": cls._SAS_TOKEN or cls._ACCOUNT_KEY,
        }

    def __init__(
        self,
        *paths: str,
        container_name: str = None,
    ):
        if container_name is None:
            assert len(paths) == 1
            path = paths[0]
            account_url = self.get_account_info()["account_url"]
            assert path.startswith(account_url)
            path = path[len(account_url) :]
            k = path.find("/")
            if k < 0:
                container_name = path
                paths = ("/",)
            else:
                container_name = path[:k]
                paths = (path[k:],)

        super().__init__(*paths)

        self._container_name = container_name

        self._container_client: ContainerClient = None
        self._blob_client: BlobClient = None
        # self._a_container_client: Optional[aContainerClient] = None
        # self._a_blob_client: Optional[aBlobClient] = None
        self._lease: BlobLeaseClient = None
        self._lock_count: int = 0

    def __getstate__(self):
        return self._container_name, super().__getstate__()

    def __setstate__(self, data):
        self._container_name, z1 = data
        self._container_client = None
        self._blob_client = None
        self._lease = None
        self._lock_count = 0
        super().__setstate__(z1)

    def __repr__(self) -> str:
        return "{}('{}', container_name='{}')".format(
            self.__class__.__name__, self._path, self._container_name
        )

    def __str__(self) -> str:
        return f"{self._container_name}://{self._path.lstrip('/')}"

    @overrides
    def as_uri(self) -> str:
        # TODO: what's the conventional lead word for Azure?
        return f"azure://{self._container_name}/{self._path.lstrip('/')}"

    @property
    def container_name(self):
        return self._container_name

    @overrides
    def is_file(self) -> bool:
        with self._provide_blob_client():
            return self._blob_client.exists()

    @overrides
    def file_info(self):
        try:
            with self._provide_blob_client():
                info = self._blob_client.get_blob_properties()
                return FileInfo(
                    ctime=info.creation_time.timestamp(),
                    mtime=info.last_modified.timestamp(),
                    time_created=info.creation_time,
                    time_modified=info.last_modified,
                    size=info.size,
                    details=info,
                )
                # If an existing file is written to again using
                # `write_...(..., overwrite=True)`,
                # then its `ctime` will not change; only `mtime`
                # will be updated.
        except ResourceNotFoundError:
            return None

    @property
    @overrides
    def root(self) -> AzureBlobUpath:
        """
        Return a new path representing the root of the same container.
        """
        return self.__class__(
            container_name=self._container_name,
        )

    def _copy_file_from(self, source, *, overwrite=False):
        # TODO: use `overwrite`
        with self._provide_blob_client():
            with source._provide_blob_client():
                copy = self._blob_client.start_copy_from_url(
                    source._blob_client.url,
                    requires_sync=True,
                )
                assert copy["copy_status"] == "success"

    @overrides
    def _copy_file(self, target: Upath, *, overwrite=False):
        if isinstance(target, AzureBlobUpath):
            target._copy_file_from(self, overwrite=overwrite)
        else:
            super()._copy_file(target, overwrite=overwrite)

    @overrides
    def download_file(self, target: LocalPathType, *, overwrite=False) -> None:
        target = _resolve_local_path(target)
        if target.is_file():
            if not overwrite:
                raise FileExistsError(str(target))
            target.remove_file()
        elif target.is_dir():
            raise IsADirectoryError(str(target))

        with self._provide_blob_client():
            # TODO: check behavior of `download_blob` about
            # overwrite.
            os.makedirs(str(target.parent), exist_ok=True)
            with open(str(target), "wb") as f:
                data = self._blob_client.download_blob()
                data.readinto(f)

    @overrides
    def upload_file(self, source: LocalPathType, *, overwrite: bool = False):
        source = _resolve_local_path(source)
        if self.is_file():
            if not overwrite:
                # TODO: check the behavior of `upload_blob` related to
                # behavior about overwrite.
                raise FileExistsError(self)
            self.remove_file()
        with self._provide_blob_client():
            with open(str(source), "rb") as data:
                self._blob_client.upload_blob(data)

    @overrides
    def iterdir(self) -> Iterator[Self]:
        with self._provide_container_client():
            prefix = self.blob_name + "/"
            k = len(prefix)
            for p in self._container_client.walk_blobs(name_starts_with=prefix):
                yield self / p.name[k:]

    def _acquire_lease(self, timeout: int = None):
        if timeout is None:
            timeout = 300
        t0 = time.perf_counter()
        while True:
            try:
                self.write_text(datetime.utcnow().isoformat(), overwrite=True)
                try:
                    self._lease = self._blob_client.acquire_lease(
                        lease_duration=-1, timeout=10
                    )
                    return
                except ResourceNotFoundError:
                    continue  # go to the outer looper to write the file again
                except HttpResponseError as e:
                    if (
                        e.status_code == 409 and e.error_code == "LeaseAlreadyPresent"
                    ):  # pylint: disable=no-member
                        # Having a lease held by others. Continue to wait.
                        # This may happen when another client placed the lease
                        # on this blob right after we've created it, that is,
                        # another worker's won out in `acquire_lease`.
                        pass
                    else:
                        raise
            except HttpResponseError as e:
                if (
                    e.status_code == 412 and e.error_code == "LeaseIdMissing"
                ):  # pylint: disable=no-member
                    # Blob exists and has a lease on it. Wait and try again.
                    pass
                else:
                    raise

            t1 = time.perf_counter()
            if t1 - t0 >= timeout:
                raise LockAcquireError(self, t1 - t0)
            time.sleep(random.uniform(0.05, 1.0))

    @contextmanager
    @overrides
    def lock(self, *, timeout=None):
        """
        References:
        https://docs.microsoft.com/en-us/azure/storage/blobs/concurrency-manage?tabs=dotnet
        """
        with self._provide_blob_client():
            if self._lease is None:
                self._acquire_lease(timeout)
                self._lock_count = 1
            else:
                self._lock_count += 1
            try:
                yield
            finally:
                self._lock_count -= 1
                if self._lock_count <= 0:
                    try:
                        try:
                            self.remove_file()
                        except FileNotFoundError:
                            pass
                        self._lease.release()
                        self._lease = None
                        self._lock_count = 0
                    except Exception as e:
                        raise LockReleaseError(
                            f"failed to release lock file {self}"
                        ) from e

    # @asynccontextmanager
    # async def a_lock(self, *, timeout=None):
    #     loop = asyncio.get_running_loop()
    #     with self._provide_blob_client():
    #         # TODO: this context manager is sync

    #         if self._lease is None:
    #             ff = functools.partial(self._acquire_lease, timeout=timeout)
    #             await loop.run_in_executor(None, ff)
    #             self._lock_count = 1
    #         else:
    #             self._lock_count += 1
    #         try:
    #             yield
    #         finally:
    #             self._lock_count -= 1
    #             if self._lock_count <= 0:
    #                 await loop.run_in_executor(None, self._lease.release)
    #                 self._lease = None
    #                 self._lock_count = 0

    @contextmanager
    def _provide_blob_client(self):
        if self._blob_client is None:
            bc = BlobClient(
                container_name=self._container_name,
                blob_name=self.blob_name,
                **self.get_account_info(),
            )
            self._blob_client = bc
            try:
                with bc:
                    yield
            finally:
                self._blob_client = None
        else:
            yield

    # TODO: how to optimize this part so that
    # new objects can reuse the ContainerClient?

    @contextmanager
    def _provide_container_client(self):
        if self._container_client is None:
            cc = ContainerClient(
                container_name=self._container_name,
                **self.get_account_info(),
            )
            self._container_client = cc
            try:
                with cc:
                    yield
            finally:
                self._container_client = None
        else:
            yield

    @overrides
    def read_bytes(self) -> bytes:
        with self._provide_blob_client():
            try:
                return self._blob_client.download_blob().readall()
            except ResourceNotFoundError as e:
                raise FileNotFoundError(self) from e

    @overrides
    def remove_file(self):
        with self._provide_blob_client():
            try:

                self._blob_client.delete_blob(
                    delete_snapshots="include", lease=self._lease
                )
            except ResourceNotFoundError:
                raise FileNotFoundError(self)

    @overrides
    def riterdir(self) -> Iterator[Self]:
        with self._provide_container_client():
            prefix = self.blob_name + "/"
            k = len(prefix)
            for p in self._container_client.list_blobs(name_starts_with=prefix):
                yield self / p.name[k:]

    @overrides
    def write_bytes(self, data: bytes, *, overwrite=False) -> None:
        if self._path == "/":
            raise UnsupportedOperation("can not write to root as a blob", self)

        with self._provide_blob_client():
            try:
                self._blob_client.upload_blob(
                    data, overwrite=overwrite, lease=self._lease
                )
            except ResourceExistsError as e:
                raise FileExistsError(self) from e



from __future__ import annotations

import contextlib
import datetime
import os
import os.path
import pathlib
import shutil
import sys
import time
import warnings
from collections.abc import Iterator
from typing import Optional, Union

import filelock

# `filelock` is also called `py-filelock`.
# Tried `fasteners` also. In one use case,
# `filelock` worked whereas `fasteners.InterprocessLock` failed.
#
# Other options to look into include
# `oslo.concurrency`, `pylocker`, `portalocker`.
from deprecation import deprecated
from overrides import overrides

from ._upath import FileInfo, LockAcquireError, LockReleaseError, Upath

# End user may want to do this:
# logging.getLogger("filelock").setLevel(logging.WARNING)


# Hack `filelock` to replace `time.nomotonic` by `time.perf_counter`
def _acquire(
    self,
    timeout: float | None = None,
    poll_interval: float = 0.05,
    *,
    poll_intervall: float | None = None,
    blocking: bool = True,
) -> filelock.AcquireReturnProxy:
    """
    Try to acquire the file lock.
    :param timeout: maximum wait time for acquiring the lock, ``None`` means use the default :attr:`~timeout` is and
        if ``timeout < 0``, there is no timeout and this method will block until the lock could be acquired
    :param poll_interval: interval of trying to acquire the lock file
    :param poll_intervall: deprecated, kept for backwards compatibility, use ``poll_interval`` instead
    :param blocking: defaults to True. If False, function will return immediately if it cannot obtain a lock on the
        first attempt. Otherwise this method will block until the timeout expires or the lock is acquired.
    :raises Timeout: if fails to acquire lock within the timeout period
    :return: a context object that will unlock the file when the context is exited
    .. code-block:: python
        # You can use this method in the context manager (recommended)
        with lock.acquire():
            pass
        # Or use an equivalent try-finally construct:
        lock.acquire()
        try:
            pass
        finally:
            lock.release()
    .. versionchanged:: 2.0.0
        This method returns now a *proxy* object instead of *self*,
        so that it can be used in a with statement without side effects.
    """
    # Use the default timeout, if no timeout is provided.
    if timeout is None:
        timeout = self.timeout

    if poll_intervall is not None:
        msg = "use poll_interval instead of poll_intervall"
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        poll_interval = poll_intervall

    # Increment the number right at the beginning. We can still undo it, if something fails.
    with self._thread_lock:
        self._lock_counter += 1

    lock_id = id(self)
    lock_filename = self._lock_file
    start_time = time.perf_counter()
    try:
        while True:
            with self._thread_lock:
                if not self.is_locked:
                    filelock._api._LOGGER.debug(
                        "Attempting to acquire lock %s on %s", lock_id, lock_filename
                    )
                    self._acquire()

            if self.is_locked:
                filelock._api._LOGGER.debug(
                    "Lock %s acquired on %s", lock_id, lock_filename
                )
                break
            elif blocking is False:
                filelock._api._LOGGER.debug(
                    "Failed to immediately acquire lock %s on %s",
                    lock_id,
                    lock_filename,
                )
                raise filelock.Timeout(self._lock_file)
            elif 0 <= timeout < time.perf_counter() - start_time:
                filelock._api._LOGGER.debug(
                    "Timeout on acquiring lock %s on %s", lock_id, lock_filename
                )
                raise filelock.Timeout(self._lock_file)
            else:
                msg = "Lock %s not acquired on %s, waiting %s seconds ..."
                filelock._api._LOGGER.debug(msg, lock_id, lock_filename, poll_interval)
                time.sleep(poll_interval)
    except BaseException:  # Something did go wrong, so decrement the counter.
        with self._thread_lock:
            self._lock_counter = max(0, self._lock_counter - 1)
        raise
    return filelock.AcquireReturnProxy(lock=self)


# `filelock` is pinned to 3.9.0 with this hack.
# Once a new `filelock` version is out, check whether
# it has made similar changes.
filelock.BaseFileLock.acquire = _acquire


class LocalUpath(Upath, os.PathLike):
    _LOCK_POLL_INTERVAL_SECONDS = 0.03

    def __init__(self, *pathsegments: str):
        """
        Create a path on the local file system.
        Both POSIX and Windows platforms are supported.

        ``*pathsegments`` specify the path, either absolute or relative to the current
        working directory. If missing, the constructed path is the current working directory.
        This is passed to `pathlib.Path <https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_.
        """
        super().__init__(str(pathlib.Path(*pathsegments).absolute()))
        self._lock_count: int = 0
        self._lock = None

    def __fspath__(self) -> str:
        """
        LocalUpath implements the `os.PathLike <https://docs.python.org/3/library/os.html#os.PathLike>`_ protocol,
        hence a LocalUpath object can be used anywhere an object implementing
        os.PathLike is accepted. For example, used with the builtin function
        `open() <https://docs.python.org/3/library/functions.html#open>`_:

        >>> p = LocalUpath('/tmp/test/data.txt')
        >>> p.rmrf()
        0
        >>> p.write_text('abc')
        >>> with open(p) as file:
        ...     print(file.read())
        abc
        >>> p.rmrf()
        1
        """
        return self.path.__fspath__()

    def __getstate__(self):
        return None, super().__getstate__()

    def __setstate__(self, data):
        _, z1 = data
        self._lock_count = 0
        self._lock = None
        super().__setstate__(z1)

    @property
    @overrides
    def path(self) -> pathlib.Path:
        """
        Return the `pathlib.Path <https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_ object
        of the path.
        """
        return pathlib.Path(self._path)

    @property
    @deprecated(
        deprecated_in="0.6.9", removed_in="0.8.0", details="Use `path` instead."
    )
    def localpath(self):
        return self.path

    @overrides
    def as_uri(self) -> str:
        """
        Represent the path as a file URI.
        On Linux, this is like 'file:///home/username/path/to/file'.
        On Windows, this is like 'file:///C:/Users/username/path/to/file'.
        """
        return self.path.as_uri()

    @overrides
    def is_dir(self) -> bool:
        """
        Return whether the current path is a dir.
        """
        return self.path.is_dir()

    @overrides
    def is_file(self) -> bool:
        """
        Return whether the current path is a file.
        """
        return self.path.is_file()

    @overrides
    def file_info(self) -> Optional[FileInfo]:
        """
        Return file info if the current path is a file;
        otherwise return ``None``.
        """
        if not self.is_file():
            return None
        st = self.path.stat()
        return FileInfo(
            ctime=st.st_ctime,
            mtime=st.st_mtime,
            time_created=datetime.datetime.fromtimestamp(st.st_ctime),
            time_modified=datetime.datetime.fromtimestamp(st.st_mtime),
            size=st.st_size,
            details=st,
        )
        # If an existing file is written to again using `write_...`,
        # then its `ctime` and `mtime` are both updated.
        # My experiments showed that `ctime` and `mtime` are equal.

    @property
    @overrides
    def root(self) -> LocalUpath:
        """
        Return a new path representing the root.

        On Windows, this is the root on the same drive, like ``LocalUpath('C:\')``.
        On Linux and Mac, this is ``LocalUpath('/')``.
        """
        return self.__class__(self.path.root)

    @overrides
    def read_bytes(self) -> bytes:
        """
        Read the content of the current file as bytes.
        """
        try:
            return self.path.read_bytes()
        except (IsADirectoryError, FileNotFoundError) as e:
            raise FileNotFoundError(self) from e

    @overrides
    def write_bytes(self, data: bytes, *, overwrite: bool = False) -> None:
        """
        Write the bytes ``data`` to the current file.
        """
        if self.is_file():
            if not overwrite:
                raise FileExistsError(self)
        self.parent.path.mkdir(exist_ok=True, parents=True)
        self.path.write_bytes(data)

        # If `self` is an existing directory, will raise `IsADirectoryError`.
        # If `self` is an existing file, will overwrite.

    @overrides
    def _copy_file(self, target: Upath, *, overwrite: bool = False):
        if isinstance(target, LocalUpath):
            if not overwrite and target.is_file():
                raise FileExistsError(target)
            os.makedirs(target.parent, exist_ok=True)
            # If `p` is a file and we try to `os.makedirs(p / 'subdir`)`,
            # on Linux it raises `NotADirectoryError`;
            # on Windows it raises `FileNotFoundError`.
            shutil.copyfile(self.path, target.path)
            # If target already exists, it will be overwritten.
        else:
            super()._copy_file(target, overwrite=overwrite)

    @overrides
    def remove_dir(self, **kwargs) -> int:
        """
        Remove the current dir along with all its contents recursively.
        """
        n = super().remove_dir(**kwargs)
        if self.path.is_dir():
            shutil.rmtree(self.path)
        return n

    @overrides
    def remove_file(self) -> None:
        """Remove the current file."""
        try:
            self.path.unlink()
        except PermissionError as e:  # this happens on Windows if `self` is a dir.
            if self.is_dir():
                raise IsADirectoryError(self) from e
            else:
                raise
        # On Linux, if `self` is a dir, `IsADirectoryError` will be raised.

    def rename_dir(
        self,
        target: str | LocalUpath,
        *,
        overwrite: bool = False,
        quiet: bool = False,
    ) -> LocalUpath:
        """Rename the current dir (i.e. ``self``) to ``target``.

        ``overwrite`` is applied file-wise. If there are
        files under ``target`` that do not have counterparts under ``self``,
        they are left untouched.

        ``quiet`` controls whether to print progress info.

        Return the new path.
        """

        if not self.is_dir():
            raise FileNotFoundError(str(self))

        if isinstance(target, LocalUpath):
            target = target._path
        target_ = self.parent / target
        if target_ == self:
            return self

        if not quiet:
            print(f"Renaming {self!r} to {target_!r}", file=sys.stderr)
        self._copy_dir(self, target_, "rename_file", overwrite=overwrite, quiet=quiet)

        def _remove_empty_dir(path):
            k = 0
            for p in path.iterdir():
                if p.is_dir():
                    k += _remove_empty_dir(p)
                else:
                    k += 1
            if k == 0:
                path.rmdir()
            return k

        _remove_empty_dir(self.path)

        return target_

    def _rename_file(self, target: str, *, overwrite=False):
        target = self.parent / target
        if not overwrite and target.is_file():
            raise FileExistsError(target)
        os.makedirs(target.parent, exist_ok=True)
        self.path.rename(target.path)

    def rename_file(
        self, target: str | LocalUpath, *, overwrite: bool = False
    ) -> LocalUpath:
        """Rename the current file (i.e. ``self``) to ``target`` in the same store.

        ``target`` is either absolute or relative to ``self.parent``.
        For example, if ``self`` is '/a/b/c/d.txt', then
        ``target='e.txt'`` means '/a/b/c/e.txt'.

        If ``overwrite`` is ``False`` (the default) and the target file exists,
        ``FileExistsError`` is raised.

        Return the new path.
        """
        if isinstance(target, LocalUpath):
            target = target._path
        target_ = self.parent / target
        if target_ == self:
            return self

        self._rename_file(target_._path, overwrite=overwrite)
        return target_

    @overrides
    def iterdir(self) -> Iterator[LocalUpath]:
        """
        Yield the immediate children under the current dir.
        """
        try:
            for p in self.path.iterdir():
                yield self / p.name
        except (NotADirectoryError, FileNotFoundError):
            pass

    @overrides
    def riterdir(self) -> Iterator[LocalUpath]:
        """
        Yield all files under the current dir recursively.
        """
        for p in self.iterdir():
            if p.is_file():
                yield p
            elif p.is_dir():
                yield from p.riterdir()

    @contextlib.contextmanager
    @overrides
    def lock(self, *, timeout=None):
        """
        This uses the package `filelock <https://github.com/tox-dev/py-filelock>`_ to implement
        a file lock for inter-process communication.
        """
        if timeout is None:
            timeout = 60
        if self._lock_count == 0:
            os.makedirs(self.parent, exist_ok=True)
            lock = filelock.FileLock(str(self))
            t0 = time.perf_counter()
            try:
                lock.acquire(
                    timeout=timeout, poll_interval=self._LOCK_POLL_INTERVAL_SECONDS
                )
            except Exception as e:
                raise LockAcquireError(
                    f"waited on '{self}' for {time.perf_counter() - t0:.2f} seconds"
                ) from e
            self._lock = lock
        self._lock_count += 1
        try:
            yield
        finally:
            self._lock_count -= 1
            if self._lock_count == 0:
                try:
                    try:
                        self.remove_file()  # This must be done before releasing the lock, or lock can "leak".
                    except FileNotFoundError:
                        pass
                    self._lock.release(force=True)
                    self._lock = None
                except Exception as e:
                    raise LockReleaseError(f"failed to release lock file {self}") from e


LocalPathType = Union[str, pathlib.Path, LocalUpath]



"""
The package *upathlib*
defines a unified API for cloud blob store (aka "object store") as well as local file systems.

Attention is focused on identifying the *most essential* functionalities
while working with a blob store for data processing.
Functionalities in a traditional local file system that are secondary in these tasks---such
as symbolic links, fine-grained permissions, and various access modes---are ignored.

End user should look to the class :class:`~upathlib.Upath` for documentation of the API.
Local file system is implemented by :class:`LocalUpath`, which subclasses Upath.
Client for Google Cloud Storage (i.e. blob store on GCP) is implemented by another Upath subclass,
namely :class:`~upathlib.GcsBlobUpath`.

One use case is the package `biglist <https://biglist.readthedocs.io/en/latest/>`__,
where the class `Biglist <https://biglist.readthedocs.io/en/latest/#biglist.Biglist>`__ takes a Upath object to indicate its location of storage.
It does not care whether the storage is local or in a cloud blob store---it
simply uses the common API to operate the storage.
"""

__version__ = "0.7.2"

from pathlib import Path
from typing import Union

from ._blob import BlobUpath
from ._local import LocalUpath
from ._upath import FileInfo, LockAcquireError, LockReleaseError, Upath

try:
    from .gcs import GcsBlobUpath
except ImportError:
    pass
try:
    from .azure import AzureBlobUpath
except ImportError:
    pass


PathType = Union[str, Path, Upath]


def resolve_path(path: PathType):
    if isinstance(path, str):
        if path.startswith("gs://"):
            # If you encounter a "gs://..." path but
            # you haven't installed GCS dependencies,
            # you'll get an exception!
            return GcsBlobUpath(path)
        if path.startswith("s3://"):
            raise NotImplementedError("AWS S3 storage is not implemented")
        if path.startswith("https://"):
            if "blob.core.windows.net" in path:
                return AzureBlobUpath(path)
            raise ValueError(path)
        path = Path(path)
    if isinstance(path, Path):
        return LocalUpath(str(path.resolve().absolute()))
    if not isinstance(path, Upath):
        raise TypeError(f"`{type(path)}` is provided while `{PathType}` is expected")
    return path
